# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nbox', 'nbox.framework', 'nbox.hyperloop', 'nbox.nbxlib', 'nbox.sub_utils']

package_data = \
{'': ['*'], 'nbox': ['assets/*'], 'nbox.framework': ['protos/*']}

install_requires = \
['Jinja2==3.0.3',
 'dill==0.3.4',
 'grpcio==1.43.0',
 'mypy-protobuf==3.2.0',
 'python-json-logger==2.0.2',
 'randomname>=0.1.3,<0.2.0',
 'requests>=2.25.1,<3.0.0',
 'tabulate==0.8.9']

setup_kwargs = {
    'name': 'nbox',
    'version': '0.9.5a0',
    'description': 'ML Inference 🥶',
    'long_description': '<a href="https://nimblebox.ai/" target="_blank"><img src="./assets/built_at_nbx.svg" align="right"></a>\n\n# 🏖️ Nbox\n\npypi `nbox` is stable for SSH and other APIs. Check out [v1](https://github.com/NimbleBoxAI/nbox/tree/v1) for the latest code and APIs!\n\nA library that makes using a host of models provided by the opensource community a lot more easier. \n\n> The entire purpose of this package is to make using models 🥶.\n\n```\npip install nbox\n```\n\n#### Current LoC\n\n```\nSLOC\tDirectory\tSLOC-by-Language (Sorted)\n996     top_dir         python=996\n88      framework       python=88\n\nTotals grouped by language (dominant language first):\npython:        1084 (100.00%)\n```\n\n## 🔥 Usage\n\n```python\nimport nbox\n\n# As all these models come from the popular frameworks you use such as \n# torchvision, efficient_pytorch or hf.transformers\nmodel = nbox.load("torchvision/mobilenetv2", pretrained = True)\n\n# nbox makes inference the priority so you can use it\n# pass it a list for batch inference \nout_single = model(\'cat.jpg\')\nout = model([Image.open(\'cat.jpg\'), np.array(Image.open(\'cat.jpg\'))])\ntuple(out.shape) == (2, 1000)\n\n# deploy the model to a managed kubernetes cluster on NimbleBox.ai\nurl_endpoint, nbx_api_key = model.deploy()\n\n# or load a cloud infer model and use seamlessly\nmodel = nbox.load(\n  model_key_or_url = url_endpoint,\n  nbx_api_key = nbx_api_key,\n  category = "image"\n)\n\n# Deja-Vu!\nout_single = model(\'cat.jpg\')\nout = model([Image.open(\'cat.jpg\'), np.array(Image.open(\'cat.jpg\'))])\ntuple(out.shape) == (2, 1000)\n```\n\n## ⚙️ CLI\n\nJust add this to your dockerfile or github actions.\n\n```\nNBX_AUTH=1 python -m nbox deploy --model_path=my/model.onnx --deployment_type="nbox"\n\n# or for more details\n\npython -m nbox --help\n```\n\n## ✏️ Things for Repo\n\n- Using [`poetry`](https://python-poetry.org/) for proper package management as @cshubhamrao says.\n  - Add new packages with `poetry add <name>`. Do not add `torch`, `tensorflow` and others, useless burden to manage those.\n  - When pushing to pypi just do `poetry build && poetry publish` this manages all the things around\n- Install `pytest` and then run `pytest tests/ -v`.\n- Using `black` for formatting, VSCode to the moon.\n- To make the docs:\n  ```\n  # from current folder\n  sphinx-apidoc -o docs/source/ ./nbox/ -M -e\n  cd docs && make html\n  cd ../build/html && python3 -m http.server 80\n  ```\n\n# 🧩 License\n\nThe code in thist repo is licensed as [BSD 3-Clause](./LICENSE). Please check for individual repositories for licenses.\n',
    'author': 'NBX Research',
    'author_email': 'research@nimblebox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NimbleBoxAI/nbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
I started with the initial assumption that abstracting away the details of
processing the data was doable but unfortunately I was not able to get it to
work reliably. We redid the entire thing by manually defining the code and
operations but it was still super hard. So the answer was to generate code
that the user will have to implement.

It loosely works like this, let's just say I am trying to write code for
``torch_to_onnx`` module. ``ModelSpec`` expects that ``source`` and ``target``
contain all the required information for running this model in another
location. Now the actual code is only of two lines, but the many different
frameworks, with their different functions and different inputs means that
any scope of manual process is gone out the window. So we basically create a
registry that can store all the information and pass it along to ``nbox.Model``.

The implementation of the registry is a ``@register`` decorator, which takes
in all the user information and underneath structures the entire thing and
stores it, if needed to be written down for compilation.

Read the code for best understanding.
"""

import os

from .. import utils as U
from ..utils import isthere, logger
from .autogen import ml_register
from .model_spec_pb2 import ModelSpec, Tensor

class NboxOptions:
  def __init__(self, model_name: str = None, folder: str = None, create_folder = False):
    model_name = model_name or U.get_random_name().replace("-", "_")
    if create_folder:
      self.folder = U.join(folder if folder != None else U.NBOX_HOME_DIR, model_name)
      self.model_name = model_name
      os.makedirs(self.folder, exist_ok = False)
    else:
      self.folder = folder
      self.model_name = model_name

@ml_register.conditional(framework = "torch",)
def torch_condition(model):
  import torch
  if isinstance(model, torch.nn.Module):
    return True
  return False

@ml_register.register(
  framework = "torch",
  target = "torch",
  stub_name = "torch_to_torchscript",
  message_name = "TorchToTorchscript",
  target_processor_name = "torchscript_to_torch",
  export_fn_import = "from torch.jit import trace",
  dependencies = ["-f https://download.pytorch.org/whl/cpu/torch_stable.html", "torch"],
)
def torch_export_torchscript(user_options, nbox_options, spec: ModelSpec):
  import torch
  from torch.jit import trace
  traced_model = trace(**user_options.__dict__)
  filepath = U.join(nbox_options.folder, "model.pt")
  spec.source.path = filepath
  spec.inputs.extend([
    Tensor(
      name = str(i),
      shape = tuple(t.shape),
      dtype = str(t.numpy().dtype)
    ) for i, t in enumerate(user_options.example_inputs)
  ])

  # actual export
  torch.jit.save(traced_model, filepath)
  return spec

@ml_register.register(
  framework = "torch",
  target = "torch",
  stub_name = "torchscript_to_torch",
  message_name = "TorchscriptToTorch",
  export_fn_import = "from torch.jit import load",
  dependencies = ["-f https://download.pytorch.org/whl/cpu/torch_stable.html", "torch"],
  ignore_args= ["f"],
)
def torch_import_torchscript(user_options, nbox_options, spec: ModelSpec):
  from torch.jit import load
  model = load(U.join(nbox_options.folder, "model.pt"), map_location="cpu")
  return model

@ml_register.register(
  framework = "torch",
  target = "onnx",
  target_processor_name = "onnx_runtime",
  stub_name = "torch_to_onnx",
  message_name="TorchToOnnx",
  export_fn_import="from torch.onnx import export",
  dependencies=["-f https://download.pytorch.org/whl/cpu/torch_stable.html", "torch"],
  ignore_args = ["f"],
)
def torch_export_onnx(user_options, nbox_options, spec: ModelSpec):
  import torch
  from torch.onnx import export
  user_options.f = nbox_options.filepath # override the folder
  inputs = user_options.args
  if isinstance(inputs, torch.Tensor):
    spec.inputs.append(Tensor(name = '0', shape = tuple(inputs.shape), dtype = str(inputs.numpy().dtype)))
  elif isinstance(inputs, tuple):
    if any(isinstance(x, dict) for x in inputs):
      if not inputs[-1] == {}:
        logger.warning(f"Last input is not an empty dictionary and is recommended by ONNX Export")
    for i,t in enumerate(inputs):
      if isinstance(t, torch.Tensor):
        spec.inputs.append(Tensor(name = str(i), shape = tuple(t.shape), dtype = str(t.numpy().dtype)))
      elif isinstance(t, dict):
        for k,v in t.items():
          spec.inputs.append(Tensor(name = k, shape = tuple(v.shape), dtype = str(v.numpy().dtype)))

  # actual export
  export(**user_options.__dict__)
  return spec


################################################################################
# User Guide
# ==========
# This 
################################################################################


@isthere("torch")
def _get_torch_model():
  import torch
  class Feedforward(torch.nn.Module):
    def __init__(self, input_size, hidden_size):
      super(Feedforward, self).__init__()
      self.input_size = input_size
      self.hidden_size = hidden_size
      self.network = torch.nn.Sequential(
        torch.nn.Linear(self.input_size, self.hidden_size),
        torch.nn.Softmax(-1),
        torch.nn.Linear(self.hidden_size, 1),
        torch.nn.Sigmoid(),
      )

    def forward(self, x):
      output = self.network(x)
      return output

  import random

  _dim = random.choice(range(10))
  return (
    {
      "model": Feedforward(input_size = _dim, hidden_size = 2)
    }, # nbox.Model kwargs
    torch.randn(1, _dim), # example input
  )

@isthere("sklearn", "numpy")
def _get_sklearn_model():
  from sklearn import linear_model
  reg = linear_model.LinearRegression()
  data = [[0, 0], [1, 1], [2, 2]], [0, 1, 2]

  return (
    {
      "model": reg,
      "method": "fit",
    }, # nbox.Model kwargs
    data, # example input
  )

def _get_default_forward():
  class SomeCaller:
    def __init__(self):
      pass

    def __call__(self):
      print("Add your logic here")

  return (
    {
      "model": SomeCaller(),
    }, # nbox.Model kwargs
    None, # example input
  )




class InvalidProtocolError(Exception):
  pass

# class ModelOutput(U.DBase):
#   __slots__ = [
#     "inputs", # :Any
#     "outputs", # :Any
#   ]

#   def serialise():
#     raise NotImplementedError

# # class ModelSpec(DBase):
# #   __slots__ = [
# #     # where from
# #     "src_framework", # :str: name of the source framework
# #     "src_framework_version", # :str
# #     'export_path', # str: there is no reason for the Pod to know anything about the user

# #     # where to
# #     "export_type", # :str
# #     "exported_time", # :str: UTC time when the model was exported

# #     # how to
# #     "load_method", # :str: The classmethod to call to load this model
# #     "load_kwargs", # :dict: kwargs to pass to the load method
# #     "io_dict", # :dict: obtained from above function

# #     # 'what to' is the serving script!
# #     "required_packages", # :list: list of things to write in requirements.txt
# #   ]

# class FrameworkAgnosticProtocol(object):
#   """
#   This Protocol only aims to serve as an abstracted layer that calls the underlying framework
#   specific code and returns the output. This mainly is intended to implement three functions:
#   * forward: the forward pass of the model
#   * export: that takes in a string for export format and arguments for it
#   * deserialise: that takes the model spec object and returns the loader for this model
#   https://nimblebox.notion.site/nbox-FrameworkAgnosticProtocol-6b39249316b1497b8ad9ff8f02b227f0
#   """
#   # def __init__(self, i0: Any, i1: Any) -> None

#   def forward(self, input_object) -> ModelOutput:
#     out = self._model(**input_object)
#     return ModelOutput(inputs = input_object, outputs = out)

#   # def export(self, format: str, input_object: Any, export_model_path: str, **kwargs) -> ModelSpec:
#   #   raise InvalidProtocolError(f"export method not implemented: {self.__class__.__name__} -> {format}")

#   def serialise(self, format):
#     if format not in self.serialisations:
#       raise InvalidProtocolError(f"Format {format} not supported for torch export")
#     return self.serialisations[format]

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:
#     raise InvalidProtocolError(f"deserialise method not implemented: '{model_meta.export_type}'")


# def get_io_dict(input_object, call_fn, forward_fn):
#   """Generates and returns an io_dict by performing forward pass through the model

#   Args:
#       input_object (Any): Input to a model
#       call_fn (Callable): function that Model.model employs to do a forward pass.
#       forward_fn (Callable): forward() function of the Model

#   Returns:
#       io : io_dict
#   """
#   logger.info(f"Generating io_dict")
#   out = forward_fn(input_object)
#   args = inspect.getfullargspec(call_fn)
#   args.args.remove("self")

#   # Create the i/o dict
#   def __get_struct(object):

#     def process_dict(x, curr_idx) -> dict:
#       #Process Dictionaries
#       results = {}
#       for key in x.keys():
#         parsed, curr_idx = parse(x[key], name=key, curr_idx=curr_idx)
#         results[key] = parsed
#       return results, curr_idx

#     def process_container(x, curr_idx) -> dict:
#       #Handles lists, sets and tuples
#       results = []
#       for element in x:
#         parsed, curr_idx = parse(element, None, curr_idx)
#         results.append(parsed)

#       return results, curr_idx

#     def parse(x, name=None, curr_idx=0) -> dict:
#       # Parses objects to generate iodict
#       dtype = None
#       if name is None:
#         name = f"tensor{curr_idx}"
#         curr_idx += 1

#       if hasattr(x, "dtype"):
#         dtype = str(x.dtype)

#       if isinstance(x, dict):
#         return process_dict(x, curr_idx)

#       elif isinstance(x, (list, set, tuple)):
#         return process_container(x, curr_idx)

#       elif hasattr(x, "shape"):
#         dim_names=[""]*len(x.shape)
#         if hasattr(x, "names"):
#           dim_names=x.names
#         return {"name": name, "dtype": dtype, "tensorShape": {"dim":[{'name':dim_names[y], "size":x.shape[y]} for y in range(len(x.shape))],"unknownRank":False}}, curr_idx
#       else:
#         return {"name": name, "dtype": dtype, "shape": None}, curr_idx

#     return parse(object)[0]

#   io = {
#     "inputs": __get_struct(out.inputs),
#     "outputs": __get_struct(out.outputs),
#     "arg_spec": {
#       "args": args.args,
#       "varargs": args.varargs,
#       "varkw": args.varkw,
#       "defaults": args.defaults,
#       "kwonlyargs": args.kwonlyargs,
#       "kwonlydefaults": args.kwonlydefaults,
#     }
#   }
#   return io


# ################################################################################
# # NimbleBox.ai Deployments
# # ========================
# # NBX-Deploy is a service that you can use to load the API endpoints where it is
# # deployed. This is a special kind of method because it can consume all the
# # objects that other models can consume.
# ################################################################################

# # TODO:@yashbonde
# # class NBXModel(FrameworkAgnosticProtocol):
# #   @U.isthere("numpy", soft=False)
# #   def __init__(self, url_key: Tuple[str, str]):
# #     logger.debug(f"Trying to load as NBX-Api")
# #     if not (isinstance(url_key, (tuple, list)) and len(url_key) == 2):
# #       raise InvalidProtocolError(f"Invalid url_key for NBX-Api")

    

# #     self.url = url
# #     self.key = key

# #     # fetch the metadata from the cloud
# #     model_url = url.rstrip("/")
# #     logger.debug("Getting model metadata")
# #     URL = secret.get("nbx_url")
# #     r = requests.get(f"{URL}/api/model/get_model_meta", params=f"url={model_url}&key={key}")
# #     try:
# #       r.raise_for_status()
# #     except:
# #       raise ValueError(f"Could not fetch metadata, please check status: {r.status_code}")

# #     # start getting the metadata, note that we have completely dropped using OVMS meta and instead use nbox_meta
# #     content = json.loads(r.content.decode())["meta"]
# #     nbox_meta = content["nbox_meta"]

# #     all_inputs = nbox_meta["metadata"]["inputs"]
# #     templates = {}
# #     for node, meta in all_inputs.items():
# #       templates[node] = [int(x["size"]) for x in meta["tensorShape"]["dim"]]
# #     logger.debug("Cloud infer metadata obtained")

# #     category = nbox_meta["spec"]["category"]

# #     # if category is "text" or if it is dict then any key is "text"
# #     tokenizer = None
# #     max_len = None
# #     if category == "text" or (isinstance(category, dict) and any([x == "text" for x in category.values()])):
# #       import transformers

# #       model_key = nbox_meta["spec"]["model_key"].split("::")[0].split("transformers/")[-1]
# #       tokenizer = transformers.AutoTokenizer.from_pretrained(model_key)
# #       max_len = templates["input_ids"][-1]

# #     image_parser = ImageParser(cloud_infer=True, post_proc_fn=lambda x: x.tolist(), templates=templates)
# #     text_parser = TextParser(tokenizer=tokenizer, max_len=max_len, post_proc_fn=lambda x: x.tolist())

# #   def forward(self, model_input):
# #     import numpy as np

# #     logger.debug(f"Hitting API: {self.model_or_model_url}")
# #     st = time()
# #     # OVMS has :predict endpoint and nbox has /predict
# #     _p = "/" if "export_type" in self.nbox_meta["spec"] else ":"
# #     json = {"inputs": model_input}
# #     if "export_type" in self.nbox_meta["spec"]:
# #       json["method"] = method
# #     r = requests.post(self.url + f"/{_p}predict", json=json, headers={"NBX-KEY": self.key})
# #     et = time() - st
# #     out = None

# #     try:
# #       r.raise_for_status()
# #       out = r.json()

# #       # first try outputs is a key and we can just get the structure from the list
# #       if isinstance(out["outputs"], dict):
# #         out = {k: np.array(v) for k, v in r.json()["outputs"].items()}
# #       elif isinstance(out["outputs"], list):
# #         out = np.array(out["outputs"])
# #       else:
# #         raise ValueError(f"Outputs must be a dict or list, got {type(out['outputs'])}")
# #       logger.debug(f"Took {et:.3f} seconds!")
# #     except Exception as e:
# #       logger.debug(f"Failed: {str(e)} | {r.content.decode()}")
# #       raise e


# ################################################################################
# # Torch
# # =====
# # Torch is a deep learning framework from Facebook. It is aimed towards a simple
# # interface and is a huge inspiration to me. User needs to define the
# # config/init and forward pass. This is the minimum input you need to take from
# # a user to get all the output
# ################################################################################




# #   @staticmethod
# #   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:
# #     import torch
# #     logger.debug(f"Deserialising torch model from {model_meta.export_path}")

# #     kwargs = model_meta.load_kwargs

# #     if model_meta.export_type == "torchscript":
# #       lp = kwargs.pop("logic_path")
# #       logger.debug(f"Loading logic from {lp}")
# #       model = torch.jit.load(model_meta.load_kwargs["model"], map_location=kwargs["map_location"])
# #       return model
# #     else:
# #       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")


# ################################################################################
# # ONNX Runtime
# # ============
# # OnnxRuntime is a C++ library that allows you to run ONNX models in your
# # application. It is a wrapper around the ONNX API. It *does not*
# # work on Apple M1 machines and so suck it up! But this is useful for server
# # side processing.
# ################################################################################

# class ONNXRtModel(FrameworkAgnosticProtocol):
#   @U.isthere("onnx", "onnxruntime", "numpy", soft = False)
#   def __init__(self, m0):
#     import onnx
#     import onnxruntime as ort
#     logger.debug(f"Trying to load as ONNX Runtime")

#     if not isinstance(m0, onnx.onnx_ml_pb2.ModelProto):
#       raise InvalidProtocolError(f"First input must be a ONNX model, got: {type(m0)}")

#     self._sess = ort.InferenceSession(self._model.SerializeToString())

#   def forward(self, input_object) -> ModelOutput:
#     # import numpy as np
#     # model_inputs = self._logic(input_object)
#     # input_names = [i.name for i in self._sess.get_inputs()]
#     # output_names = [o.name for o in self._sess.get_outputs()]
#     # if isinstance(model_inputs,dict):
#     #   tensor_dict = model_inputs
#     #   if isinstance(list(model_inputs.values())[0],SklearnInput):
#     #     tensor_dict = {list(model_inputs.keys())[0]:list(model_inputs.values())[0].inputs}
#     #   model_inputs = {name: tensor for name, tensor in zip(input_names, tuple(tensor_dict.values()))}
#     # else:
#     #   raise ValueError("Your Logic needs to return dictionary")

#     # # conversion to proper arrays -> if can be numpy-ed and float32
#     # for key, value in model_inputs.items():
#     #   if hasattr(value, "numpy"):
#     #     value = value.numpy()
#     #   if value.dtype in (np.float64, np.float32, np.float16):
#     #     value = value.astype(np.float32)
#     #   elif value.dtype in (np.int64, np.int32, np.int16, np.int8, np.int64):
#     #     value = value.astype(np.int32)
#     #   model_inputs[key] = value

#     # res = self._sess.run(output_names=output_names, input_feed=model_inputs)

#     res = self._sess.run(**input_object)
#     return ModelOutput(
#       inputs = input_object,
#       outputs = res,
#     )

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Any:
#     import onnx
#     logger.debug(f"Deserialising ONNX model from {model_meta.export_path}")

#     if model_meta.export_type == "onnx":
#       model = onnx.load(model_meta.load_kwargs["model"])
#       return model
#     else:
#       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")


# ################################################################################
# # Sklearn
# # =======
# # Scikit-learn is a Python library that allows you to use classical machine
# # learning algorithms in your application. Its versatility is huge and that
# # means that a dedicate Input DataModel is required.
# ################################################################################

# class SklearnModel(FrameworkAgnosticProtocol):
#   @U.isthere("sklearn", "numpy", soft = False)
#   def __init__(self, m1):
#     logger.debug(f"Trying to load as Sklearn")
#     if not "sklearn" in str(type(m1)):
#       raise InvalidProtocolError

#     self._model = m1

#     self.exporters = {
#       "pkl": U.to_pickle
#     }
#     self.loaders = {
#       "pkl": U.from_pickle
#     }

#   def forward(self, input_object: Dict) -> ModelOutput:
#     """Sklearn covers a far wider range of models and is therefore more complicated to cover.
#     For now the input is expected to be a Dict that has "inputs" dict that will be fed to the model.
#     And it has "method" key that will be called upon the model i.e. whether user wants
#     ``predict`` (default) or ``predict_proba`` is left upto the user.
#     """

#     # if "sklearn.neighbors.NearestNeighbors" in str(type(self._model)):
#     #   method = getattr(self._model, "kneighbors") if method == None else getattr(self._model, method)
#     # elif "sklearn.cluster" in str(type(self._model)):
#     #   if any(
#     #     x in str(type(self._model)) for x in ["AgglomerativeClustering", "DBSCAN", "OPTICS", "SpectralClustering"]
#     #   ):
#     #     method = getattr(self._model, "fit_predict")
#     # else:
#     #   try:
#     #     method = getattr(self._model, "predict") if method == None else getattr(self._model, method)
#     #   except Exception as e:
#     #     logger.debug(f"[ERROR] Model Prediction Function is not yet registered {e}")

#     model_input = input_object.get("inputs", None)
#     method = input_object.get("method", "predict")
#     method_fn = getattr(self._model, method, None)
#     if method_fn == None:
#       logger.error(f"Method {method} not found in SklearModel")
#       raise ValueError(f"Method {method} not found in model")
#     out = method(**model_input)

#     return ModelOutput(inputs = model_input, outputs = out,)

#   @U.isthere("skl2onnx", soft = False)
#   def export_to_onnx(self, input_object, export_model_path, model_file_name="model.onnx", opset_version=None,) -> ModelSpec:
#     import skl2onnx
#     import skl2onnx.common.data_types as dt

#     __NP_DTYPE_TO_SKL_DTYPE = {
#       "bool_": dt.BooleanTensorType,
#       "double": dt.DoubleTensorType,
#       "int16": dt.Int16TensorType,
#       "complex128": dt.Complex128TensorType,
#       "int32": dt.Int32TensorType,
#       "uint16": dt.UInt16TensorType,
#       "complex64": dt.Complex64TensorType,
#       "float16": dt.Float16TensorType,
#       "int64": dt.Int64TensorType,
#       "str_": dt.StringTensorType,
#       "uint32": dt.UInt32TensorType,
#       "float64": dt.FloatTensorType,
#       "uint64": dt.UInt64TensorType,
#       "int8": dt.Int8TensorType,
#       "uint8": dt.UInt8TensorType,
#     }

#     export_path = join(export_model_path, model_file_name)
#     method = input_object.get("method", "predict")
#     method = getattr(self._model, method, None)
#     if method == None:
#       logger.error(f"Method {method} not found in SklearModel")
#       raise ValueError(f"Method {method} not found in model")
#     iod = get_io_dict(input_object, method, self.forward)

#     # the args need to be converted to proper formatted data types
#     initial_types = []
#     input_names, input_shapes, input_dtypes = [],[],[]

#     # Case for input object with single tensor
#     inputs=iod["inputs"]
#     if isinstance(inputs['name'], str):
#       input_names.append(inputs["name"])
#       input_dtypes.append(inputs["dtype"])
#       tensorShape = list()
#       for i in inputs["tensorShape"]["dim"]:
#         tensorShape.append(i["size"])
#       input_shapes.append(tensorShape)
#     else:
#       for key in inputs:
#         input_names.append(inputs[key]['name'])
#         input_dtypes.append(inputs[key]['dtype'][key])
#         tensorShape = list()
#         for i in inputs[key]["tensorShape"]["dim"]:
#           tensorShape.append(i["size"])
#         input_shapes.append(tensorShape)

#     for name, shape, dtype in zip(input_names, input_shapes, input_dtypes):
#       shape[0] = None # batching requires the first dimension to be None
#       initial_types.append((name, __NP_DTYPE_TO_SKL_DTYPE[str(dtype)](shape)))
#     onx_model = skl2onnx.to_onnx(self._model, initial_types=initial_types, target_opset=opset_version)

#     with open(export_path, "wb") as f:
#       f.write(onx_model.SerializeToString())

#     return ModelSpec(
#       export_type="onnx",
#       export_path=export_path,
#       load_class=self.__class__.__name__,
#       load_method="from_to_onnx",
#       load_kwargs={
#         "model": f"./{model_file_name}",
#         "map_location": "cpu",
#       },
#       io_dict=iod,
#     )

#   def export_to_pkl(self, input_object, export_model_path, model_file_name = "model.pkl"):
#     export_path = join(export_model_path, model_file_name)
#     with open(export_path, "wb") as f:
#       dill.dump(self._model, f)

#     method = input_object.get("method", None)
#     method = getattr(self._model, "predict") if method == None else getattr(self._model, method)
#     iod = get_io_dict(input_object, method, self.forward)

#     return ModelSpec(
#       export_type = "pkl",
#       load_kwargs = {
#         "model": f"./{model_file_name}",
#         "map_location": "cpu",
#       },
#       io_dict = iod,
#     )

#   def export(self, format, input_object, export_model_path, **kwargs) -> ModelSpec:
#     if format == "onnx":
#       model_spec = self.export_to_onnx(input_object, export_model_path=export_model_path,**kwargs)
#     elif format == "pkl":
#       model_spec = self.export_to_pkl(input_object, export_model_path=export_model_path, **kwargs)
#     else:
#       raise InvalidProtocolError(f"Unsupported export format for torch Module: {format}")

#     import sklearn
#     model_spec.src_framework = "sklearn"
#     model_spec.src_framework_version = sklearn.__version__

#     return model_spec

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:
#     logger.info(f"Deserialising SkLearn model")
#     kwargs = model_meta.load_kwargs
#     if model_meta.export_type == "pkl":
#       with open(kwargs["model"], "rb") as f:
#         model = dill.load(f)

#     else:
#       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")

#     return model, logic


# ################################################################################
# # Tensorflow
# # ==========
# # Tensorflow is a high-level machine learning library built by Google. It is
# # part of a much larger ecosystem called tensorflow extended.
# ################################################################################

# class TensorflowModel(FrameworkAgnosticProtocol):
#   @U.isthere("tensorflow", soft = False)
#   def __init__(self, m0):
#     import tensorflow as tf
#     logger.debug(f"Trying to load Tensorflow model")
#     if not isinstance(m0, (tf.keras.Model, tf.Module)):
#       raise InvalidProtocolError(f"First input must be a tensorflow module or Keras model, got: {type(m0)}")

#     self._model = m0

#   def forward(self, input_object: Any) -> ModelOutput:
#     model_inputs = self._logic(input_object)
#     if isinstance(model_inputs, dict):
#       out = self._model(list(model_inputs.values()))
#     else:
#       out = self._model(model_inputs)
#     return ModelOutput(inputs = input_object, outputs = out)

#   def _serialise_logic(self, fpath):
#     logger.debug(f"Saving logic to {fpath}")
#     with open(fpath, "wb") as f:
#       dill.dump(self._logic, f)

#   @U.isthere("tf2onnx", soft = False)
#   def export_to_onnx(
#     self,
#     input_object,
#     export_model_path,
#     logic_file_name="logic.dill",
#     model_file_name="model.onnx",
#     opset_version=None,
#   ) -> ModelSpec:
#     import tf2onnx
#     import tensorflow as tf

#     input_object = self._logic(input_object)

#     self._serialise_logic(join(export_model_path, logic_file_name))
#     export_path = join(export_model_path, model_file_name)
#     iod = get_io_dict(input_object, self._model.call, self.forward)

#     if "keras" in str(self._model.__class__):
#       model_proto, _ = tf2onnx.convert.from_keras(self._model, opset=opset_version)
#       with open(export_path, "wb") as f:
#         f.write(model_proto.SerializeToString())

#       return ModelSpec(
#         src_framework="tf",
#         src_framework_version=tf.__version__,
#         export_type="onnx",
#         export_path=export_path,
#         load_class=self.__class__.__name__,
#         load_method="from_convert_from_keras",
#         load_kwargs={
#           "model": f"./{model_file_name}",
#           "map_location": "cpu",
#           "logic_path": f"./{logic_file_name}",
#         },
#         io_dict=iod,
#       )
#     else:
#       raise InvalidProtocolError(f"Unsupported ONNX export for tensorflow model: {type(self._model)}")

#   def export_to_savemodel(
#     self,
#     input_object,
#     export_model_path,
#     logic_file_name = "logic.dill",
#     model_dir_name = "model",
#     include_optimizer = True,
#   ) -> ModelSpec:
#     import tensorflow as tf

#     input_object = self._logic(input_object)
#     self._serialise_logic(join(export_model_path, logic_file_name))

#     export_path = join(export_model_path, model_dir_name)
#     iod = get_io_dict(input_object, self._model.call, self.forward)

#     if "keras" in str(self._model.__class__):
#       tf.keras.models.save_model(
#         self._model,
#         export_path,
#         signatures = None,
#         options = None,
#         include_optimizer = include_optimizer,
#         save_format = "tf"
#       )

#       return ModelSpec(
#         src_framework = "tf",
#         src_framework_version = tf.__version__,
#         export_type = "SaveModel",
#         export_path = export_path,
#         load_class = self.__class__.__name__,
#         load_kwargs = {
#           "model": f"./{model_dir_name}",
#           "map_location": "cpu",
#           "logic_path": f"./{logic_file_name}"
#         },
#         io_dict = iod,
#       )
#     else:
#       raise InvalidProtocolError(f"Unsupported SaveModel export for tensorflow model: {type(self._model)}")

#   def export_to_h5(
#     self,
#     input_object,
#     export_model_path,
#     logic_file_name = "logic.dill",
#     model_file_name = "model.h5py",
#     include_optimizer = True,
#   ) -> ModelSpec:

#     import tensorflow as tf

#     self._serialise_logic(join(export_model_path, logic_file_name))
#     iod = get_io_dict(input_object, self._model.call, self.forward)

#     export_path = join(export_model_path, model_file_name)

#     if "keras" in str(self._model.__class__):
#       tf.keras.models.save_model(
#         self._model,
#         export_path,
#         signatures=None,
#         options=None,
#         include_optimizer=include_optimizer,
#         save_format = "h5"
#       )

#       return ModelSpec(
#         src_framework = "tf",
#         src_framework_version = tf.__version__,
#         export_type = "h5",
#         export_path = export_path,
#         load_class = self.__class__.__name__,
#         load_method = "from_savemodel",
#         load_kwargs = {
#           "model": f"./{model_file_name}",
#           "map_location": "cpu",
#           "logic_path": f"./{logic_file_name}"
#         },
#         io_dict = iod,
#       )
#     else:
#       raise InvalidProtocolError(f"Unsupported h5 export for tensorflow model: {type(self._model)}")

#   def export(self, format: str, input_object: Any, export_model_path: str, **kwargs) -> ModelSpec:
#     if format == "onnx":
#       return self.export_to_onnx(
#         input_object, export_model_path = export_model_path, **kwargs
#       )
#     elif format == "SaveModel":
#       return self.export_to_savemodel(
#         input_object, export_model_path=export_model_path, **kwargs
#       )
#     elif format == "h5":
#       return self.export_to_h5(
#         input_object, export_model_path=export_model_path, **kwargs
#       )
#     else:
#       raise  InvalidProtocolError(f"Unknown format: {format}")

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:

#     logger.info(f"Deserialising Tensorflow model from {model_meta.export_path}")
#     kwargs = model_meta.load_kwargs

#     if model_meta.export_type in ["SaveModel", "h5"]:
#       lp = kwargs.pop("logic_path")
#       logger.info(f"Loading logic from {lp}")
#       with open(lp, "rb") as f:
#         logic = dill.load(f)

#       import tensorflow as tf
#       model = tf.keras.models.load_model(model_meta.load_kwargs["model"])
#       return model, logic
#     else:
#       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")

# ################################################################################
# # Jax
# # ===
# # Jax is built by Google for high performance ML research. Read more here:
# # https://jax.readthedocs.io/en/latest/notebooks/quickstart.html
# ################################################################################

# """ We are not implementing Jax as of now"""


# ################################################################################
# # Flax
# # ===
# # Flax is built by Google, it is a neural network library and ecosystem for JAX
# # designed for flexibility. Read more here: https://github.com/google/flax
# ################################################################################

# class FlaxModel(FrameworkAgnosticProtocol):
#   @U.isthere("flax", soft = False)
#   def __init__(self, m0):
#     import flax
#     from flax import linen as nn

#     logger.debug("Trying to load Flax model")

#     if not isinstance(m0, nn.Module):
#       raise InvalidProtocolError(f"First input must be a Flax module, got: {type(m0)}")

#     if not isinstance(m1, flax.core.frozen_dict.FrozenDict):
#       # m1 should contain the model parameters
#       raise ValueError(f"Second input for flax model must be a\
#        FrozenDict containing model parameters, got: {type(m1)}")

#     self._model = m0
#     self._params = m1

#   def forward(self, input_object: Any) -> ModelOutput:
#     model_inputs = input_object
#     if type(model_inputs) is dict:
#       out = self._model.apply(self._params, **model_inputs)
#     else:
#       out = self._model.apply(self._params, model_inputs)

#     return ModelOutput(inputs=input_object, outputs=out)

#   def _serialise_params(self, fpath):
#     logger.info(f"Saving parameters to {fpath}")
#     raise NotImplementedError

#   def export(self, format: str, input_object: Any, export_model_path: str, **kwargs) -> ModelSpec:
#    raise NotImplementedError

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:
#     raise NotImplementedError



"""
This file provides support for the SSH tunneling into NBX-Build Instances.
"""
# If you have come till here, and you want to work on more things like this, reach out:
# research@nimblebox.ai

import os
import ssl
import socket
import certifi
import threading
from typing import List
from datetime import datetime
from functools import partial

from ..utils import NBOX_HOME_DIR, logger as nbx_logger
from .. import utils as U
from ..auth import secret
from ..instance import Instance

class FileLogger:
  def __init__(self, filepath):
    self.filepath = filepath
    self.f = open(filepath, "w")

    self.debug = partial(self.log, level="debug",)
    self.info = partial(self.log, level="info",)
    self.warning = partial(self.log, level="warning",)
    self.error = partial(self.log, level="error",)
    self.critical = partial(self.log, level="critical",)

  def log(self, message, level):
    self.f.write(f"[{datetime.utcnow().isoformat()}] {level}: {message}\n")
    self.f.flush()


class RSockClient:
  """
  This is a RSockClient. It handels the client socket where client is the user application trying to connect to "client_port"
  Connects to RSockServer listening on localhost:886.
  RSockServer recieves instructions as a string and sends back a response.
  RSockServer requires following steps to setup
  First, authentication:
    - Authentication happens by sending
      `"AUTH~{AUTH_TOKEN}"`
    - AUTH_TOKEN is not defined and is default to 'password'
  Seting config:
    - you can set config by sending
      `"SET_CONFIG~{instance}~{instance_port}"`
    - "instance" - Currently is the internal ip of the instance.
    - "instance_port" - What port users wants to connect to.
  Connect:
    - This Starts the main loop which
      1. Listen on client_port
      2. On connection, 
        a. Send AUTH
        b. If AUTH is successful, send SET_CONFIG
        c. If SET_CONFIG is successful, send CONNECT
        d. If CONNECT is successful, start io_copy
  IO_COPY:
    - This is the main loop that handles the data transfer between client and server. This is done by creating a new thread for each connection.
    - The thread is created by calling the function "io_copy" for each connection that is "server" and "client".
    - When a connection is closed, the loop is stopped.
  """

  def __init__(self, connection_id, client_socket, instance, instance_port, auth, logger: FileLogger, secure=False,):
    """
    Initializes the client.
    Args:
      client_socket: The socket that the client is connected to.
      instance: The instance that the client wants to connect to.
      instance_port: The port that the instance is listening on.
      auth: The authentication token that the client has to provide to connect to the RSockServer.
      secure: Whether or not the client is using SSL.
    
    """
    self.connection_id = connection_id
    self.client_socket = client_socket
    self.instance = instance
    self.instance_port = instance_port
    self.auth = auth
    self.secure = secure
    self.logger = logger

    self.client_auth = False
    self.rsock_thread_running = False
    self.client_thread_running = False

    self.logger.info('Starting client')
    self.connect_to_rsock_server()
    self.logger.info('Connected to RSockServer')
    self.authenticate()
    self.logger.info('Authenticated client')
    self.set_config()
    self.logger.info('Client init complete')

  def connect_to_rsock_server(self):
    """
    Connects to RSockServer.
    """
    self.logger.debug('Connecting to RSockServer')
    rsock_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    rsock_socket.connect(('rsocks.nimblebox.ai', 886))

    if self.secure:
      self.logger.info('Starting SSL')
      self.rsock_socket = ssl.wrap_socket(rsock_socket, ca_certs=certifi.where(), cert_reqs=ssl.CERT_REQUIRED)
    else:
      self.rsock_socket = rsock_socket

  def authenticate(self):
    """
    Authenticates the client.
    Sends `"AUTH~{AUTH_TOKEN}"` to RSockServer.
    """
    self.logger.info('Authenticating client')
    self.rsock_socket.sendall(bytes(f'AUTH~{self.auth}', 'utf-8'))
    auth = self.rsock_socket.recv(1024)
    auth = auth.decode('utf-8')
    if auth == 'OK':
      self.logger.info('Client authenticated')
    else:
      self.logger.erro('Client authentication failed')
      self.client_auth = False
      exit(1)
  
  def set_config(self):
    """
    Sets the config of the client.
    Sends `"SET_CONFIG~{instance}~{instance_port}"` to RSockServer.
    """
    self.logger.info('Setting config')
    self.rsock_socket.sendall(bytes(f'SET_CLIENT~{self.instance}~{self.instance_port}', 'utf-8'))
    config = self.rsock_socket.recv(1024)
    config = config.decode('utf-8')
    self.logger.info('Config set to {}'.format(config))
    if config == 'OK':
      self.logger.info('Config set')
    else:
      self.logger.error('Config set failed')
      exit(1)

  def connect(self):
    """
    Connects the client to RSockServer.
    Sends `"CONNECT"` to RSockServer.
    """
    self.logger.info('Starting the io_copy loop')
    self.rsock_socket.sendall(bytes('CONNECT', 'utf-8'))
    
    # start the io_copy loop
    self.rsock_thread_running = True
    self.rsock_thread = threading.Thread(target=self.io_copy, args=("server", ))
    self.rsock_thread.start()

    # start the io_copy loop
    self.client_thread_running = True
    self.client_thread = threading.Thread(target=self.io_copy, args=("client", ))
    self.client_thread.start()

  def io_copy(self, direction):
    """
    This is the main loop that handles the data transfer between client and server.
    """
    self.logger.info('Starting {} io_copy'.format(direction))

    if direction == 'client':
      client_socket = self.client_socket
      server_socket = self.rsock_socket

    elif direction == 'server':
      client_socket = self.rsock_socket
      server_socket = self.client_socket

    while True:
      try:
        data = client_socket.recv(1024)
        if data:
          # self.logger.info('{} data: {}'.format(direction, data))
          server_socket.sendall(data)
        else:
          self.logger.info('{} connection closed'.format(direction))
          break
      except Exception as e:
        self.logger.error('Error in {} io_copy: {}'.format(direction, e))
        break
    self.logger.info('Stopping {} io_copy'.format(direction))

def create_connection(local_port, instance_id, instance_port, logger: FileLogger, listen = 1):
  listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  listen_socket.bind(('localhost', local_port))
  listen_socket.listen(listen)

  connection_id = 0

  while True:
    logger.info('Waiting for client')
    client_socket, _ = listen_socket.accept()
    logger.info('Client connected')

    connection_id += 1
    logger.info(f'Total clients connected -> {connection_id}')
    # create the client
    pwd = secret.get("access_token")

    client = RSockClient(
      connection_id = connection_id,
      client_socket = client_socket,
      instance = instance_id,
      instance_port = instance_port,
      auth = pwd,
      logger = logger,
      secure = True
    )

    # start the client
    client.connect()

def port_in_use(port: int) -> bool:
  import socket
  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    return s.connect_ex(('localhost', port)) == 0

def tunnel(ssh: int, *apps_to_ports: List[str], i: str):
  """the nbox way to SSH into your instance, by default ``"jupyter": 8888 and "mlflow": 5000``

  Usage:
    tunn.py 8000 notebook:8000 2000:8001 -i "nbox-dev"

  Args:
    ssh: Local port to connect SSH to
    *apps_to_ports: A tuple of values ``<app_name/instance_port>:<localport>``.
      For example, ``jupyter:8888`` or ``2001:8002``
    i(str): The instance to connect to
    pwd (str): password to connect to that instance.
  """

  import sys

  if sys.platform.startswith("linux"):  # could be "linux", "linux2", "linux3", ...
    pass
  elif sys.platform == "darwin":
    pass
  elif sys.platform == "win32":
    # Windows (either 32-bit or 64-bit)
    raise Exception("Windows is unsupported platform, raise issue: https://github.com/NimbleBoxAI/nbox/issues")
  else:
    raise Exception(f"Unkwown platform '{sys.platform}', raise issue: https://github.com/NimbleBoxAI/nbox/issues")

  folder = U.join(NBOX_HOME_DIR, "tunnel_logs")
  os.makedirs(folder, exist_ok=True)
  filepath = U.join(folder, f"tunnel_{i}.log")
  file_logger = FileLogger(filepath)

  
  # ===============

  default_ports = {
    "jupyter": 8888,
    "mlflow": 5000,
  }
  apps = {} # <localport-cloudport>
  for ap in apps_to_ports:
    app, port = ap.split(':')
    port = int(port)
    if not app or not port:
      raise ValueError(f"Invalid app:port pair {ap}")
    try:
      apps[int(app)] = port
    except:
      if app not in default_ports:
        raise ValueError(f"Unknown '{app}' should be either integer or one of {', ' .join(default_ports.keys())}")
      apps[port] = default_ports[app]
  apps[ssh] = 22 # hard code

  ports_used = []
  for k in apps:
    if port_in_use(k):
      ports_used.append(str(k))

  if ports_used:
    raise ValueError(f"Ports {', '.join(ports_used)} are already in use")

  # check if instance is the correct one
  instance = Instance(i)
  if not instance.state == "RUNNING":
    raise ValueError("Instance is not running")
  nbx_logger.info(f"password: {instance.open_data['ssh_pass']}")

  # create the connection
  threads = []
  for local_port, cloud_port in apps.items():
    nbx_logger.info(f"Creating connection from {cloud_port} -> {local_port}")
    t = threading.Thread(target=create_connection, args=(local_port, instance.instance_id, cloud_port, file_logger, 1))
    t.start()
    threads.append(t)

  try:
    # start the ssh connection on terminal
    import subprocess
    nbx_logger.info(f"Starting SSH ... for graceful exit press Ctrl+D then Ctrl+C")
    subprocess.call(f'ssh -p {ssh} ubuntu@localhost', shell=True)
  except KeyboardInterrupt:
    nbx_logger.info("KeyboardInterrupt, closing connections")
    for t in threads:
      t.join()

  sys.exit(0) # graceful exit
